#include "ehs/system/Semaphore_W32.h"
#include "ehs/UTF.h"
#include "ehs/Log.h"

namespace ehs
{
	Semaphore::~Semaphore()
    {
        if (!hdl)
			return;

		if (!CloseHandle(hdl))
		{
			DWORD code = GetLastError();
			EHS_LOG_INT("Error", 0, Str_8("Failed to free semaphore with error #") + code + ".");
		}
    }

	Semaphore::Semaphore()
		: hdl(nullptr)
    {
		Semaphore::Initialize();
    }

	Semaphore::Semaphore(Str_8 name, const UInt_32 initial)
		: BaseSemaphore(std::move(name), initial), hdl(nullptr)
    {
		Semaphore::Initialize();
    }

	Semaphore::Semaphore(const UInt_32 initial)
		: BaseSemaphore(initial), hdl(nullptr)
    {
		Semaphore::Initialize();
    }

	Semaphore::Semaphore(Semaphore&& sem) noexcept
		: BaseSemaphore(std::move(sem)), hdl(sem.hdl)
	{
		sem.hdl = nullptr;
	}

	Semaphore::Semaphore(const Semaphore& sem)
		: BaseSemaphore(sem), hdl(nullptr)
	{
	}

	Semaphore& Semaphore::operator=(Semaphore&& sem) noexcept
	{
		if (this == &sem)
			return *this;

		Release();

		BaseSemaphore::operator=(std::move(sem));

		hdl = sem.hdl;

		sem.hdl = nullptr;

		return *this;
	}

	Semaphore& Semaphore::operator=(const Semaphore& sem)
	{
		if (this == &sem)
			return *this;

		Release();

		BaseSemaphore::operator=(sem);

		hdl = nullptr;

		return *this;
	}

	void Semaphore::Initialize()
	{
		if (IsValid())
			return;

		hdl = CreateSemaphoreW(nullptr, (LONG)GetInitial(), EHS_SINT_32_MAX, UTF::To_16(GetName()));
		if (!hdl)
			EHS_LOG_INT("Error", 0, Str_8("Failed to create semaphore with error #") + GetLastError() + ".");
	}

	void Semaphore::Release()
	{
		if (!IsValid())
			return;

		if (!CloseHandle(hdl))
		{
			DWORD code = GetLastError();
			EHS_LOG_INT("Error", 0, Str_8("Failed to free semaphore with error #") + code + ".");
		}
		hdl = nullptr;
	}

    void Semaphore::Signal(const UInt_32 inc)
    {
		if (!IsValid())
			return;

        if (!ReleaseSemaphore(hdl, (LONG)inc, nullptr))
            EHS_LOG_INT("Error", 0, Str_8("Failed to release semaphore with error #") + GetLastError() + ".");
    }

    bool Semaphore::Wait(const UInt_32 timeout)
    {
		if (!IsValid())
			return false;

        DWORD result = WaitForSingleObject(hdl, timeout);
        if (result == WAIT_ABANDONED)
        {
            EHS_LOG_INT("Error", 0, "Wait abandoned.");
            return false;
        }
        else if (result == WAIT_FAILED)
        {
            EHS_LOG_INT("Error", 1, Str_8("Wait failed with error #") + GetLastError() + ".");
            return false;
        }
        else if (result == WAIT_TIMEOUT)
            return false;

        return true;
    }

	bool Semaphore::IsValid() const
	{
		return hdl;
	}
}